/*
* $Log$
* Revision 2.1  2004/02/18 19:25:05  richq
* All the battle stuff - vs random-map-nasties and
* vs other characters/garrisons.
*
*/
#include "fight.h"
#include "flames.h"
#include "characters.h"
#include "army.h"
#include "main.h"

#define FIGHTING_GARRISON 0x99

void loseHorse(int who, int chance) {
  if ( (chance & 3) != 1) 
    return;
  
  if ( (g_characters[who].race_orders & 0xf) >= 12)
    return;
  
  g_characters[who].race_orders = g_characters[who].race_orders | 1;
}


void fightResults(int fightWhat, int who, int theirChance) {
  // 7ffb
  if (g_characters[who].object == 3) {
    // runes
    return;
  }
  
  loseHorse(who, theirChance);
  
  int r = randomish();
  if (!(r&1)) {
    return;
  }
  
  if (g_characters[who].reckless >= r)
    return;
  
  if (g_characters[who].object != 0xff && g_characters[who].object >= 5) {
    // extra "throw" for carrying a battle object
    if (randomish() >= 0x50)
      return;
  }
  
  // dead
  g_characters[who].reckless = 0;
  if (g_characters[who].object != 0xff) {
    g_objects[g_characters[who].object].x = g_characters[who].location.x;
    g_objects[g_characters[who].object].y = g_characters[who].location.y;
  }
  g_characters[who].object = fightWhat;
  if (fightWhat >= 0x80) {
    // died fighting a nasty 
    return;
  }
  // killed foe, won battle
  g_characters[fightWhat].flags |= WON_BATTLE_FLAG | KILLED_FOE_FLAG;
  g_characters[fightWhat].fighting = who;
  
}

extern int getLocation ( int x, int y );
//6705
void defenderFlees(int defender, int attacker) {
  if (defender >= 0x80) // a garrison
    return; 
  
  int x = g_characters[defender].location.x;
  int y = g_characters[defender].location.y;
  x ^= (randomish()&3);
  y ^= (randomish()&3);
  int loc = getLocation(x,y)&0xf;
  int cnt = 16;
  while (loc == WASTES && cnt) {
    x = g_characters[defender].location.x ^ (randomish()&3);
    y = g_characters[defender].location.y ^ (randomish()&3);
    loc = getLocation(x,y)&0xf;
    cnt--;
  }
  if (cnt == 0 && loc == WASTES)
  {
    g_characters[defender].location.x = x;
    g_characters[defender].location.y = y;
  }

  // lost the battle, and leave the tunnel (?)
  g_characters[defender].flags &= ~(WON_BATTLE_FLAG|IN_TUNNEL_FLAG); 
  decreaseU8By(&g_characters[defender].despondence, 32);
  
  if (attacker >= 0x80) // a garrison
    return; 
  g_characters[attacker].flags |= WON_BATTLE_FLAG;
  
}


void defenderLostArmy(int defender, int attacker) {
  // 813a
  // call 6705
  defenderFlees(defender, attacker);
  
  if (defender >= 0x80) {
    g_garrisons[defender-128].owner = attacker;
  } else {
    fightResults(FIGHTING_GARRISON,defender,0);
  }
  // 814f
}


//; army strengths:
//80ad 
const u8 armyBaseStrengths[] = {
  0xdc, 0xb4, 0xdc, 0xb4, 0xdc, 0xb4, 0xc8, 0x96, 
  0xb4, 0x8c, 0xdc, 0xb4, 0xdc, 0xb4, 0xdc, 0xa0
};

// as = energy/2 + base/2, 
// garrisons have half this strength
int getAdjustedStrength(int armyid) {
  // 808f
  int r = getArmyRace(armyid);
  int e = getArmyEnergy(armyid);
  int strength = armyBaseStrengths[r]/2 + e/2;
  
  if (armyid >= 0x80)
    strength /= 2;
  
  return strength;
  
}


void battleArmies(int attacker, int defender) {
  int attackerStr = getAdjustedStrength(attacker);
  int defenderStr = getAdjustedStrength(defender);
  int attackerSize = getArmySize(attacker);
  int defenderSize = getArmySize(defender);
  attackerStr *= attackerSize;
  attackerStr *= ((randomish()&0x1f)+16);
  
  int killed = ((attackerStr/defenderStr)>>8)&0xFF;
  
  if (defenderSize < killed) {
    // call 813a
    defenderLostArmy(defender, attacker);
    killed = defenderSize;
  }
  setArmySize(defender, defenderSize - killed);
  
  if (defender < 0x80) {
    increaseU8By(&g_charArmies[defender].lost ,  killed);
  }
  if (attacker < 0x80) {
    increaseU8By(&g_charArmies[attacker].slew ,  killed);
  }
}


// 7f85
void doBattle(int attacker, int chance) {
  int value = randomish();
  int startIndex = value%g_total_here;
  int endIndex = startIndex;
  
  int defender = g_charactersHere[startIndex];
  
  
  while (g_characters[defender].reckless == 0 || 
      LOYALTY(g_characters[defender].flags) == g_status.loyalty) {
    
    startIndex++;
    if (startIndex == g_total_here)
      startIndex = 0;
    if (startIndex  == endIndex) // if true, have reached end
      return;
      
    defender = g_charactersHere[startIndex];
    
  } 
  
  //7fb7
  g_characters[defender].flags |= IN_BATTLE_FLAG; // in fight
  if (attacker < 128) {
    // attacker is not a garrison
    // chance of killing off the foe!
    if (!KILLED_FOE(g_characters[attacker].flags))
      g_characters[attacker].fighting = defender;
    int B = randomish();
    if (chance >= B) {
      // chance lord is killed- careful, attacker may be a garrison
      if (attacker < 128)
        fightResults(attacker, defender, B);
      else
        fightResults(FIGHTING_GARRISON, defender, B);
    }
    
  }
  // everyone carries on from here...
  battleArmies(attacker, defender);
  battleArmies(defender, attacker);
}


void battleVsCharacter(int who) {
  // 7f4e
  g_status.loyalty = LOYALTY(g_characters[who].flags);
  int fightChance;
  // calculate chance:
  if (g_characters[who].object != 0xff && g_characters[who].object >= 5)
    fightChance = 0xff;
  else 
    fightChance = (g_characters[who].energy/2 + g_characters[who].despondence/2)/2 + 
                  g_characters[who].reckless / 2;
  
  doBattle(who, fightChance);
  
}
