
/*
* $Log$
* Revision 2.2  2004/02/18 19:19:39  richq
* Tunnel display stuff. Draws the flames, the posts and the door/steps.
*
* Revision 2.1.1.1  2004/02/04 21:49:21  Default
* Doomdark's Revenge GBA
*
*/


#include "flames.h"
#include "features.h"
#include "location.h"
#include "stepdata.h"
#include "gfx.h"
#include "main.h"
#include "inter.h"

// an offset due to smaller resolution
#define FLAME_OFFSET 4
#define POST_COLOUR 2
#define RIGHTPOST_OFFSET 26
#define DOOR_X_OFFSET 2
#define DOOR_Y_OFFSET 2

// how often to redraw flames in 60ths of seconds
#define FLAME_RATE 2

typedef struct {
  u8 x;
  u8 y;
  u8 col;
  u8 tile;
}
flame_t;

// the flame gfx are inside the font!
const flame_t flames[] = {
  {1,12,70,91},               /* 91 = [, 92 = '\' */
  {2,12,70,123},              // 123 = {, 124 = |
  {1,13,70,93},               // 93 = ], 94 = ^
  {2,13,70,125},              // 125 = }, 126 = ~
  {0,14,70,91},
  {1,14,115,91},
  {2,14,115,123},
  {3,14,70,123},
  {0,15,70,93},
  {1,15,115,93},
  {2,15,115,125},
  {3,15,70,125},
};

typedef struct {
  u8 width;
  u8 tile;
}
post_t;

const post_t post1 = {
  4, // width
  96, // tile
};
const post_t post2 = {
  2, // width
  64, // tile
};
const post_t post3 = {
  2, // width
  96, // tile
};

void drawFlames(void);
void drawSteps(void);

//#define fontGFX _binary_font_bin_start 
//extern u8 fontGFX[];
// 98FD (22808 in the .sna)
// 20-7f = 5f = 760 bytes 

/* Draw the flames to the buffer... 
this is gonna be a problem because the tunnel gfx are hacked into DR,
i.e. they aren't drawn "properly" using the landscaping technique
*/
void drawFlameSegment(int xOffset) {
  int i;
  // 2 ways to do this. First way would be like this:
//  
//  gfxTable = fontGFX;
//  u8 tileBuffer[8*8];
//  for (i = 0; i < 12; i++) {
//    createTile( tileBuffer,
//                flames[i].tile + (rand255()&1), 
//                1 + (flames[i].col&7),
//                1 + ((flames[i].col>>3)&7)
//               );
//    // tile contains gfx data now draw it at x,y..
//    // copy it to the main gfx buffer, then 
//    // blit that as usual
//  }
//  
  // this 2nd way is slightly more elegant and makes use of the text bg
  // this required setting the text palettes equal to speccy colours.
  // oh and the flame bg colour hack is due to lack of "colour clash"
  for (i = 0; i < 12; i++) {
    print_text(flames[i].tile+(randomish()&1),
               flames[i].x + xOffset,
               flames[i].y - FLAME_OFFSET,
               (flames[i].col&7));
    if ( ((flames[i].col/8)&7) != 0 ) {
      set_flame_tile(flames[i].x + xOffset,flames[i].y - FLAME_OFFSET);
    }
    
  }
  
}

// draws horizontal section of post
void drawPostSection(const post_t * section, int yOffset, int xOffset) {
  int i;
  for (i = 0; i < section->width; i++) {
    print_text(section->tile, xOffset+i, yOffset, POST_COLOUR);
  }
  
}

// draws horizontal section of steps
void drawStepSection(const steps_t * section, int yOffset, int xOffset) {
  int i;
  for (i = 0; i < section->width; i++) {
    print_text(section->tile, xOffset+i, yOffset, section->colours[i]&7);
  }
  
}


void drawSteps(void) {
  int y;
  // door
  for (y = 0; y < 5; y++) 
    drawStepSection(stepTable[y], y+10-FLAME_OFFSET+DOOR_Y_OFFSET, stepTable[y]->x - DOOR_X_OFFSET);
  // steps down left
  for (y = 5; y < 8; y++) 
    drawStepSection(stepTable[y], y+10-FLAME_OFFSET+DOOR_Y_OFFSET, stepTable[y]->x-(y-5) - DOOR_X_OFFSET);
  
  // steps middle
  drawStepSection(stepTable[8], 18-FLAME_OFFSET+DOOR_Y_OFFSET, stepTable[8]->x - DOOR_X_OFFSET);
  
  // steps down right
  for (y = 9; y < 12; y++) 
    drawStepSection(stepTable[y], y+10-FLAME_OFFSET+DOOR_Y_OFFSET, stepTable[y]->x+(y-9) - DOOR_X_OFFSET);
  
}

void drawPosts(int xOffset) {
  
  drawPostSection(&post1, 0x10-FLAME_OFFSET, xOffset+0);
  drawPostSection(&post2, 0x11-FLAME_OFFSET, xOffset+1);
  int B;
  for (B = 0x12; B < 0x18; B++) 
    drawPostSection(&post3, B-FLAME_OFFSET, xOffset+1);
  
}

void drawFlames(void) {
  drawFlameSegment(0);
  drawFlameSegment(RIGHTPOST_OFFSET);
}

u16 g_rseed;
int randomish(void) {
  if (!g_rseed) {
    g_rseed = game_frames&0xffff;
    g_rseed = (!g_rseed)?0x3d47:g_rseed;
  }
  
  unsigned char B = (g_rseed>>7)&0xff;
  unsigned char A = B + g_rseed;
  g_rseed = (A<<8) | B;
  return A;
}

void tunnelVision(void) {
  // 6fba
  clear_view();
  if (g_status.canDrawInfront == 0)
    return;
  
  game_frames = 0;
  // flames
  // 6f4b
  drawFlames();
  // posts
  drawPosts(0);
  drawPosts(RIGHTPOST_OFFSET);
  if (g_locdat.terrain >= GATE && g_locdat.terrain <= PALACE)
    drawSteps();
  
}

/*redraw flames every FLAME_RATE 60ths of a second*/
void redrawFlames(void) {
  if (g_status.charInTunnel == 0 || 
      g_status.canDrawInfront == 0 ||
      g_status.screen != LOOK_SCREEN
      )
    return;
  
  if (game_frames > FLAME_RATE) {
    if ( (randomish()&1)) {
      drawFlameSegment(0);
    } else {
      drawFlameSegment(RIGHTPOST_OFFSET);
    }
    
    game_frames = 0;
  }
}


