
/*
* $Log$
* Revision 2.4  2006/07/30 11:44:53  rich
* map pointer set to 0 explicitly
*
* Revision 2.3  2004/02/22 15:33:26  richq
* Table to keep track of what we have seen so far. Int to s8 changes.
* Looking into the mist fixes.
*
* Revision 2.2  2004/02/18 19:17:36  richq
* Holds the map. Calculates postions we are looking at, what we can "see", sets up
* the armies and characters at a given location. The novelty for DR is the area
* tracing stuff, used in the name generation. That's in here too.
*
* Revision 2.1.1.1  2004/02/04 21:49:20  Default
* Doomdark's Revenge GBA
*
*                                        
*/

#include "location.h"
#include "stdlib.h"
#include "string.h"
#include "features.h"
#include "characters.h"
#include "objects.h"
#include "main.h"
#include "army.h"
#include "flames.h"
#include "tokens.h"
#include "tokendata.h"
#include "gfx.h"
#include "mapscreen.h"

////location_t position;
locdat_t g_locdat;
#define mapInRom _binary_terrainmap_bin_start
extern u8 mapInRom[];

#define MAP_SIZE 96*64

location_t * g_fortresses;
location_t * g_objects;
u8 * g_map = 0;

int g_plainsBetweenUs;

location_t g_inFront;
location_t g_lookAtPos;

// filled in when we check location info..
details_t g_localDetails;


const s8 positionModifier[][2] = {
      { 0, -1},      // N  ... x =  0, y = -1
      { 1, -1},      // NE ... x =  1, y = -1
      { 1,  0},      //  E ... x =  1, y =  0
      { 1,  1},      // SE ... x =  1, y =  1
      { 0,  1},      // S  ... x =  0, y =  1
      {-1,  1},      // SW ... x = -1, y =  1
      {-1,  0},      //  W ... x = -1, y =  0
      {-1, -1}       // NW ... x = -1, y = -1
};


#define FORTRESS_LOC_X              (0)
#define FORTRESS_LOC_Y              (128)


// initialisation of the map data in RAM
void init_map(void) {
  if (g_map == 0) {
    g_map = (u8*) malloc(MAP_SIZE);
    // this table keeps track of what we have seen so far:
    g_haveSeenMap = (u8*) malloc(MAP_SIZE);
    g_fortresses = (location_t*)malloc(128 * sizeof(location_t));
    g_objects = (location_t*)malloc(128 * sizeof(location_t));
    
//    gbaprintf("sizeof(location_t) %d",sizeof(location_t));
    
  }
  
  // copy from rom to ram
  memcpy(g_map, mapInRom, MAP_SIZE);
  int i;
  for (i = 0; i < 128; i++) {
    g_fortresses[i].x = tablesInRom[FORTRESS_LOC_X + i];
    g_fortresses[i].y = tablesInRom[FORTRESS_LOC_Y + i];
    g_objects[i].x = randomish()&0x3f;
    g_objects[i].y = 0x10 + (randomish()&0x3f);
  }
  // clear out the zones we have seen
  i = 0;
  while (i < MAP_SIZE)
    g_haveSeenMap[i++] = 0;
  
}


// taken from 6db6
// calculate what is in a particular square
void fillLocationData(locdat_t * data) {
  if (data->x >= 0x40 || data->y >= 0x60){
    data->terrain = WASTES;
    data->info = 0;
  } else {
    int index = data->x+(data->y)*0x40;
    data->info = g_map[index];
    data->terrain = g_map[index]&0xf;
  }
}

// project curr pos till we hit something
// 64da in ddr
void lookingAt(const location_t * pos, int dir) {
  // position is current position. 
  // use positionModifier to do this.
  
  int plains = 0;
  
  int xmod = positionModifier[dir][0];
  int ymod = positionModifier[dir][1];
  
  locdat_t loc;
  loc.x = pos->x;
  loc.y = pos->y;
  
  g_inFront.x = pos->x+xmod;
  g_inFront.y = pos->y+ymod;
  
  do {
    loc.x += xmod;
    loc.y += ymod;
    fillLocationData(&loc);
    if (loc.terrain == PLAINS && !HAS_MIST(loc.info)) {
      plains++;
      if (plains == 3) {
        break;
      }
    }
  } while (loc.terrain == PLAINS && !HAS_MIST(loc.info));
    
  g_lookAtPos.x = loc.x;
  g_lookAtPos.y = loc.y;
  g_plainsBetweenUs = plains; 
}


// 6a2e
// checks if we can see down a tunnel
// calculates what square we are looking at, 
// calculates the square immediately infront of us
void checkVision(int character) {
  
  g_status.charInTunnel = 0;
  g_status.canDrawInfront = 1;
  g_status.loyalty  = LOYALTY(g_characters[character].flags);
  g_status.looking = LOOKING(g_characters[character].time_looking);
  
  if (IN_TUNNEL(g_characters[character].flags)) {
    // 6a43
    g_status.charInTunnel = 1;
  }
  
  lookingAt(&g_characters[character].location,
            g_status.looking);
            
  // find out what's in front...
  g_locdat.x = g_inFront.x;
  g_locdat.y = g_inFront.y;
  
  fillLocationData(&g_locdat);
  g_status.terrain_infront = g_locdat.terrain;
  
  if (g_status.charInTunnel) {
    g_status.terrain_infront = WASTES;
    
    if (!(g_status.looking & 1)) {
      // looking straight, but are we looking down a tunnel?
      if (HAS_TUNNEL(g_locdat.info)) {
        // yes.
        return;
      }
    }
    // see nothing...
    g_status.canDrawInfront = 0;
  } else {
    if (g_status.terrain_infront == WASTES)
      g_status.canDrawInfront = 0;
  }
  
  
}



/*
 workOutLocationDetails Works out location info for a character
 based on code at 87f9
 
 checks vision
 creates lists of characters at location
 
 
*/
void workOutLocationDetails(int character) {
  
  
  checkVision(character);
  
  //85f0
  g_locdat.x = g_characters[character].location.x;
  g_locdat.y = g_characters[character].location.y;
  fillLocationData(&g_locdat);
  getCharactersAt(&g_characters[character].location);
  g_localDetails.garrison_here = g_garrisonHere;
  g_localDetails.unloyals_here = g_unloyals_here;
  g_localDetails.thing_here = calculateObject(&g_characters[character].location);
  // remove creature from our location
  g_map[g_characters[character].location.x + g_characters[character].location.y*64] &= ~16;
  
  
  g_locdat.x = g_inFront.x;
  g_locdat.y = g_inFront.y;
  fillLocationData(&g_locdat);
  
  getCharactersAt(&g_inFront);
  g_localDetails.garrison_infront = g_garrisonHere;
  g_localDetails.unloyals_infront = g_unloyals_here;
  g_localDetails.thing_infront = calculateObject(&g_inFront);
  g_localDetails.can_move = g_status.canDrawInfront;
  
  if (g_localDetails.unloyals_here) {
    // possible AI use only?
    g_localDetails.can_move = 0;
  }
  
  
}

int calculateNameSeed(int x, int y) {
  int key = (444*((y*64)+x))%6151;
  return key;
}

// terrain to check
int g_checkingTerrain;


//int isContiguous(int dir, location_t * pos) {
//  switch (dir) {
//    case 0: // look north
//        if (0 == pos->y) return 0;
//        pos->y--;
//        break;
//    case 1: // east
//      if (63 == pos->x) return 0;
//      pos->x++;
//      break;
//    case 2: // south
//      if (0x5f == pos->y) return 0;
//      pos->y++;
//      break;
//    default: // west
//      if (0 == pos->x)  return 0;
//      pos->x--;
//      
//  }  
//  return (g_map[pos->x+pos->y*64]&0xf) == g_checkingTerrain;  
//}

// 5fdf
// pos will change
int isContiguous(int dir, location_t * pos) {
  if (dir == 0) {
    // look north
    if (pos->y == 0) {
      return 0;
    } else {
      pos->y--;
    }
  } else if (dir == 1) {
    // east
    if (pos->x == 63) {
      return 0;
    } else {
      pos->x++;
    }
    
  } else if (dir == 2) {
    // south
    if (pos->y == 0x5f) {
      return 0;
    } else {
      pos->y++;
    }
  } else {
    // west
    if (pos->x == 0) {
      return 0;
    } else {
      pos->x--;
    }
  }
  
  return (g_map[pos->x+pos->y*64]&0xf) == g_checkingTerrain;
  
}

/* p is in/out and will change! */
void traceArea(location_t * p) {
  g_checkingTerrain = (g_map[p->x+p->y*64]&0xf);
  
  int dir;
  int haveLooked;
  do {
    dir = 0;
    haveLooked = 0;
    location_t plainsPos = {p->x, p->y};
    location_t tmp = plainsPos;
    
    // go up until we are not stood on plains
    while ( isContiguous(0,&plainsPos) ) {
      tmp.x = plainsPos.x;
      tmp.y = plainsPos.y;
    }
    // get back to last plains position:
    plainsPos.x = tmp.x;
    plainsPos.y = tmp.y;
    
    // now go east, then south then west
    for (;;) {
      dir++;
      if (dir == 4) // this means we are on a single square "island" of plains
        return;
      if (isContiguous(dir,&plainsPos)) {
        break;
      }
      // else it isn't... it is another type of terrain
      // try another direction...
      plainsPos.x = tmp.x;
      plainsPos.y = tmp.y;
    }
    
    int firstLooked = dir;
    
    // ok, we have found the direction to search in
    location_t contiguous = plainsPos;
    location_t best  = plainsPos;
    // and store initial guess at return value
    p->x = best.x; p->y = best.y;
    
    // 5f9e
    do {
      if ( best.y > contiguous.y ||
          (best.y == contiguous.y && best.x >= contiguous.x) ) 
      {
        // we have found a better position (norther, wester)
        best.x = contiguous.x;
        best.y = contiguous.y;
      }
      
      dir = (dir-1)&3;
      tmp.x = contiguous.x;
      tmp.y = contiguous.y;
      haveLooked--;
      // look around us, see where the area continues:
      while (!isContiguous(dir,&contiguous)) {
        // look in a different direction:
        dir = (dir+1)&3;
        contiguous.x = tmp.x;
        contiguous.y = tmp.y;
        haveLooked++;
      } 
      // loop until we return to our initial position
    } while (contiguous.x != p->x || contiguous.y != p->y || firstLooked != dir);
    
    p->x = best.x;
    p->y = best.y;
  } while (haveLooked != 4);
  
}

void clearArmyTiles(){
  int i;
  int index;
  for (i = 0; i < 128; i++) {
    // bit 6 is the army flag
    // if ( has army)
    if (g_charArmies[i].size) {
      index = g_characters[i].location.x + 64*g_characters[i].location.y;
      g_map[index] &= ~HAS_ARMY_FLAG;
    }
    
  }
  
}

void setArmyTiles(){
  int i;
  int index;
  for (i = 0; i < 128; i++) {
    // bit 6 is the army flag
    // if (is alive, has army, not in tunnel)
    if (g_characters[i].reckless && g_charArmies[i].size && !IN_TUNNEL(g_characters[i].flags)) {
      index = g_characters[i].location.x + 64*g_characters[i].location.y;
      g_map[index] |= HAS_ARMY_FLAG;
    }
  }
  
}




// 7a12
void getEntranceAndObject(const location_t * pos, s8 * take, u8 * enter) {

  g_locdat.x = pos->x;
  g_locdat.y = pos->y;
  fillLocationData(&g_locdat);

  *take = -1;
  *enter = 0;

  if (!g_status.charInTunnel &&
       g_locdat.terrain >= GATE &&
       g_locdat.terrain <= PALACE
  ) {
    *enter = 1;
  }

  int i;

  for (i = 0; i < 128; i++) {
    if (g_objects[i].x == pos->x &&
        g_objects[i].y == pos->y )
    {
      *take = i;
      break;
    }
  }
  calculateObject(pos);

}


void copyMistBytes(u8 * destination,
                   u8 * source,
                   int amount) {  
  do {
    *destination = (*destination&0xbf)|(*source&0x40);
    destination--;
    source--;
    amount--;
  }
  while (amount > 0);
  
}


// the mists wrap around the board in y direction
void moveMists() {
  u8 lastRow[64];
  int i;
  for (i = 0; i < 64; i++)
    lastRow[i] = g_map[0x17c0 + i];
  
  copyMistBytes(&g_map[0x17FF], &g_map[0x17BF], 0x17c0);
  copyMistBytes(&g_map[0x3f],&lastRow[0x3f], 0x40);
  
}

void putThingsOnMap() {
  int r,i;
  for (i = 0; i < (0xc000-0xa800); i++) {
    r = randomish()&3;
    if ( (g_map[i]&0xf) >= GATE  || (g_map[i]&0xf) == r ) {
      g_map[i] |= HAS_OBJECT_FLAG; // has object
    } else {
      g_map[i] &= ~HAS_OBJECT_FLAG; // no object
    }
  }
  
  
}

void describeFortress(const location_t * pos, int whereToken, int garrison) {
  
  g_locdat.x = pos->x;
  g_locdat.y = pos->y;
  fillLocationData(&g_locdat);
  if (g_locdat.terrain == FORTRESS || g_locdat.terrain == CITY) {
    // 79b2
    addNewlineToken();
    addUppercaseToken();
    addSingleToken(whereToken);
    //7738
    addTerrainToken(pos); //the <terrain> of <place>
    addSingleToken(0x75); //holds
    addArmySizeAndType(garrison); // "one hundred barbarian warriors"
    if (getArmySize(garrison)!=0) {
      // "they are not tired"
      addToTokenList(theyAre);
      addUtterly(getArmyEnergy(garrison));
      addToTokenList(tired_);
    }
    addUppercaseToken();
    //79c2
    addTerrainToken(pos); //the <terrain> of <place>
    addToTokenList(isOwnedBy);  // is owned by
    addTitleAndLoyalty(g_garrisons[garrison-128].owner);
    print_tokens_pal(FONT_RED);
  }
}
