
/*
* $Log$
* Revision 2.3  2004/02/22 15:32:50  richq
* Width of positionModifier changed from int to its actual value.
*
* Revision 2.2  2004/02/18 19:17:36  richq
* Holds the map. Calculates postions we are looking at, what we can "see", sets up
* the armies and characters at a given location. The novelty for DR is the area
* tracing stuff, used in the name generation. That's in here too.
*
* Revision 2.1.1.1  2004/02/04 21:49:20  Default
* Doomdark's Revenge GBA
*
*                                              
*/

#ifndef _LOCATION_H
#define _LOCATION_H

#include "gba.h"
#include "features.h"

extern u8 * g_map;



/*
* location data - uses more bit flags in ddr than in lom
* so use macros to get the data from locdat.info
*/
#define HAS_OBJECT_FLAG 16
#define HAS_ARMY_FLAG 32
#define HAS_MIST_FLAG 64
#define HAS_TUNNEL_FLAG 128

#define HAS_OBJECT(info) ((info&HAS_OBJECT_FLAG)!=0)
#define HAS_ARMY(info) ((info&HAS_ARMY_FLAG)!=0)
#define HAS_MIST(info) ((info&HAS_MIST_FLAG)!=0)
#define HAS_TUNNEL(info) ((info&HAS_TUNNEL_FLAG)!=0)
typedef struct Locdat {
   u8 x;        
   u8 y;        
   u8 terrain;  
   u8 info;    
} locdat_t;

typedef struct {
  u8 x;
  u8 y;
} location_t;

extern location_t * g_fortresses;
extern location_t * g_objects;

extern const s8 positionModifier[][2];

typedef struct {
  u8 unloyals_here;
  u8 unloyals_infront;
  u8 garrison_here;
  u8 garrison_infront;
  u8 thing_here;
  u8 thing_infront;
  u8 can_move;
} details_t;

extern details_t g_localDetails;

// detailed information on a location. filled in workOutLocationDetails
extern locdat_t g_locdat;
//// the square we are looking at
extern location_t g_lookAtPos;
// and how far away it is..
extern int g_plainsBetweenUs;

//// the square immediately infront of us
extern location_t g_inFront;    

void init_map(void);

// data must have its x, y vals filled!
void fillLocationData(locdat_t * data);
//void saveLocationData(void);

void lookingAt(const location_t * pos, int dir);
void workOutLocationDetails(int character);
void checkVision(int character);


// calculate the ID (take) of an object to take from this location, or -1 if none
// calculate if there is an entrance (enter) at this location, or 0 if none
void getEntranceAndObject(const location_t * pos, s8 * take, u8 * enter);

int calculateNameSeed(int x, int y);
void traceArea(location_t * p);

void clearArmyTiles(void);
void setArmyTiles(void);

void moveMists(void);
void putThingsOnMap(void);

void describeFortress(const location_t * where, int whereToken, int garrison);


#endif

