/*
*  $Log$
*  Revision 2.3  2004/02/22 15:33:59  richq
*  Set the square we have explored
*
*  Revision 2.2  2004/02/18 19:17:14  richq
*  Looking and movement. Sets up the current vista, applies movement cost when
*  moving, handles player input for moving. Similar to LoM.
*
*  Revision 2.1.1.1  2004/02/04 21:49:21  Default
*  Doomdark's Revenge GBA
*
*
*
*/
#include "gba.h"
#include "keypad.h"

#include "choose.h"
#include "infoscreen.h"
#include "look.h"
#include "savearea.h"
#include "objects.h"
#include "tokens.h"
#include "tokendata.h"
#include "features.h"
#include "gfx.h"
#include "menu.h"
#include "location.h"
#include "characters.h"
#include "main.h"
#include "flames.h"
#include "mapscreen.h"

int g_forwardsBlocked;

// 7344 ; terrain costs table
const u8 terrainBaseCost[16] = {
     1,   // plains      0x0
     8,   // mountains   0x1
     4,   // forest      0x2
     3,   // hills       0x3
     2,   // Gate        0x4
     2,   // Temple      0x5
     2,   // Pit         0x6
     2,   // Palace      0x7
     2,   // Fortress    0x8
     2,   // Hall        0x9
     2,   // Hut         0xa
     2,   // Tower       0xb
     2,   // City        0xc
     2,   // Fountain    0xd
     2,   // Stones      0xe
     2,   // Wastes      0xf
};
    

// 7354 ; race OK across terrain table
const u8 racePreferredTerrain[16] = {
     0x0c,       // LUXOR_HORSE      city
     0x0c,       // LUXOR            city
     0x00,       // MORKIN_HORSE     plains
     0x00,       // MORKIN           plains
     0x0b,       // RORTHRON_HORSE   tower
     0x0b,       // RORTHRON         tower
     0x02,       // FEY_HORSE        forest
     0x02,       // FEY              forest
     0x03,       // BARBARIAN_HORSE  hills
     0x03,       // BARBARIAN        hills
     0x00,       // ICELORD_HORSE    plains
     0x00,       // ICELORD          plains
     0x02,       // TARITHEL         forest
     0x01,       // GIANT            mountains
     0x00,       // SHARETH          plains
     0x0f        // DWARF            wastes -> "tunnels"
};

// 7364; race time taken per move table
const u8 raceMovementCost[16] = {
     3,      // LUXOR_HORSE     
     6,      // LUXOR           
     3,      // MORKIN_HORSE    
     6,      // MORKIN          
     2,      // RORTHRON_HORSE  
     4,      // RORTHRON        
     3,      // FEY_HORSE       
     6,      // FEY             
     5,      // BARBARIAN_HORSE 
     8,      // BARBARIAN       
     3,      // ICELORD_HORSE   
     5,      // ICELORD         
     2,      // TARITHEL        
     5,      // GIANT           
     2,      // SHARETH         
     5,      // DWARF           
};

void drawDisplay(void) {
  // 641d
  
  // clear top screen
  clear_text();
  remove_shield();
  // clear bottom
  clear_objects();
  
  // set horizon palette
  if (g_status.charInTunnel) {
    set_tunnel_palette();  
  }
  else {
    set_horizon();
  }
  
  drawCharGfx(g_characters[g_saveArea.selectedChar].race_orders&0xf, 
              240 - 5*8,
              7*8,
              COLOURTYPE_VIEW);
  
  createHeStandsString(g_saveArea.selectedChar, g_status.charInTunnel);
  print_tokens(0, 0, 0, 0x17, FONT_RED_B);
  
  if (g_status.charInTunnel) {
    tunnelVision();
  }
  else {
    drawPanorama(g_status.looking);  
    set_palette_time(TIME_LEFT(g_characters[g_saveArea.selectedChar].time_looking));
  }
  
  // draw things in front of us, depends on time and tunnellyness
  int ctype = COLOURTYPE_DAY;
  if (TIME_LEFT(g_characters[g_saveArea.selectedChar].time_looking) == 0)
    ctype = COLOURTYPE_NIGHT; // night
  else if (TIME_LEFT(g_characters[g_saveArea.selectedChar].time_looking) == DAWN)
    ctype = COLOURTYPE_DAWN; // dawn
  
  if (g_status.charInTunnel)
    ctype = COLOURTYPE_TUNNEL; // tunnel
    
  drawInFront(ctype);
  
}


// Look is at 844b this time
void look(void) {
  int fromScreen = g_status.screen;
  g_status.screen = LOOK_SCREEN;
  if (checkLocationAndStatus())
    return;
  // 641d
  if (fromScreen != LOOK_SCREEN) {
    fade_down();    
    clearHighlighter();
    g_panning = 0;
  }
  
  drawDisplay();
  
  // set this square as explored
  haveExplored(&g_characters[g_saveArea.selectedChar].location,
               &g_inFront);
  
  if (fromScreen != LOOK_SCREEN)
    fade_up();

}



// 729e
void applyMovementCost(int movingChar) {
  
  int movementCost = terrainBaseCost[g_status.terrain_infront];
  int charRace = g_characters[movingChar].race_orders&0xf;
  
  if (g_status.terrain_infront == racePreferredTerrain[charRace])
    movementCost = 0;
  
  movementCost += raceMovementCost[charRace];
  
  if (HAS_MIST(g_locdat.info)
    && (charRace) != RACE_ICELORD_HORSE
    && (charRace) != RACE_ICELORD
    && (charRace) != RACE_SHARETH
  ) {
    // I can buy this, more tired in mists, OK.
    // but what if we are in a tunnel, eh?
    
  
    if (!(g_status.charInTunnel && 
    (g_status.terrain_infront < GATE || g_status.terrain_infront > PALACE)
    ) ){
      // we are not in a tunnel, or moving to a misty out of tunnel area
      makeCharacterTired(movingChar);
      movementCost += 2;
      decreaseU8By(&g_characters[movingChar].despondence, 6);
    }
    
  }
  
  if (g_status.looking & 1) {
    // diagonal
    movementCost += 2;
  }
  
  if (g_localDetails.thing_infront <= OBJECT_SKULKRIN) {
    // what?? "oh what a tiring flat plain"
    // or worse, a Fey getting extra tired in the forest?
    makeCharacterTired(movingChar);
    // this looks like a bug to me... increase movement cost just for the hell of it
    movementCost += 2;
    
    // unless... this might be to encourage us to use tunnels?
  }
  if (!IS_SWIFT(g_characters[movingChar].attribute_good)) {
    movementCost++;
  }
  if (IS_SLOW(g_characters[movingChar].attribute_bad)) {
    movementCost++;
  }
  
  // decrease time left...
  u8 time = TIME_LEFT(g_characters[movingChar].time_looking);
  decreaseU8By(&time, movementCost);
  g_characters[movingChar].time_looking = (g_characters[movingChar].time_looking&7)|time<<3;
  makeCharacterTired(movingChar);
  
  if (g_status.terrain_infront == MOUNTAINS) {
    makeCharacterTired(movingChar);
    if ((charRace) != RACE_GIANT)
      makeCharacterTired(movingChar);
  }
  
}

void moveCharForwards(int charToMove) {
  applyMovementCost(charToMove);
  g_characters[charToMove].location = g_inFront;
  
  if (g_locdat.terrain >= GATE && g_locdat.terrain <= PALACE) {
    // moved out of tunnel, unset bit 6
    g_characters[charToMove].flags &= ~IN_TUNNEL_FLAG;
  }
}


// 7295 
void doMove(void) {
  
  // clear armies...
  clearArmyTiles();
  
  moveCharForwards(g_saveArea.selectedChar);
  
  // set armies
  setArmyTiles();
  
}


void moveForwards(void) {
  checkLocationAndStatus();
  
  if (!g_status.canDrawInfront) {
    return;
  }
  
  if (g_localDetails.unloyals_infront) {
    infoPlace(0);
    return;
  }
  
  if ( TIME_LEFT(g_characters[g_saveArea.selectedChar].time_looking) == 0) {
    // night time
    choose();
    return;
  }
  
  
  if ( g_characters[g_saveArea.selectedChar].energy < 0x20) {
    infoCharacter();
    return;
  }
  
  
  g_panning = 0;
  doMove();
  look();
}

void rotateLeft(void) {
  while (KeyPressed(KEY_LEFT) ) {
    int dir = g_characters[g_saveArea.selectedChar].time_looking;
    dir = (dir-1)&7;
    
    g_characters[g_saveArea.selectedChar].time_looking = MASK_LOOKING(g_characters[g_saveArea.selectedChar].time_looking) | dir;
  
    if (g_panningOn)
      g_panning = PAN_LEFT;
    
        
    look();
    if (!g_panningOn || g_status.charInTunnel)
      break;
    
    UpdateKeys();
  }
  
}
void rotateRight(void) {
  
  while (KeyPressed(KEY_RIGHT) ) {

    int dir = g_characters[g_saveArea.selectedChar].time_looking;
    dir = (dir+1)&7;
    
    g_characters[g_saveArea.selectedChar].time_looking = MASK_LOOKING(g_characters[g_saveArea.selectedChar].time_looking) | dir;
  
    if (g_panningOn) {
      g_panning = PAN_RIGHT;
    }
    look();
    if (!g_panningOn || g_status.charInTunnel)
      break;
    UpdateKeys();
  }
  
}


