/*
* $Log$
* Revision 2.4  2006/07/30 11:36:34  rich
* Char signed warning
*
* Revision 2.3  2004-02-22 15:01:08  richq
* Start row for text clearing
*
* Revision 2.2  2004/02/18 19:11:14  richq
* Very similar to the LoM token parsing, except the tokens aren't compressed into
* 5 bits and the general feel is "dumb and duller".
*  Number to words has changed, for the worse imo, as the "two score and three"
* thing is harder to follow.
*  The "utterly bold" type string is now not as complex (and always negative!)
*  Time now shows just full hours instead of the good old "less than..." thing
* Oh well...
*
* Revision 2.1.1.1  2004/02/04 21:49:20  Default
* Doomdark's Revenge GBA
*
*
*/

/*
* Everything to do with text tokens and token parsing
* See tokendata for definitions
*/

#include "keypad.h"
#include "qmath.h"

#include "tokens.h"
#include "tokendata.h"
#include "location.h"
#include "gfx.h"
#include "stdlib.h"

#include "characters.h"

//#include "memory_defs.h"

#define TOKEN_LIST_SIZE 0x100
#define TOKEN_BUFFER_SIZE 0x20

#define START_ROW 2
#define FINAL_ROW 20

// define the tokens list, 
// the list contains the tokens (8-bit values)
unsigned char * g_tokenList;// = (char*)ADDR_G_TOKENLIST;
unsigned char * g_heap;       // current position in g_tokenList

// the buffer contains the ASCII characters
char * g_tokenBuffer;// = (char*)ADDR_G_TOKENBUFFER;
char * g_bufferPos;           // current position in token buffer

// the current size of the word/token being expanded
int g_tokenLength;
// uppercase capitalises first letter, capslock capitalises whole sentences
int g_lowercase;
int g_capsLock;

// pluralness is usually based on the previous numberToWords value
int g_isPlural;

// used for text display:
int g_startX;
int g_startY;
int g_indent;   // where to start X when a newline is reached (left hand side)
int g_maxX;     // the end of the line (right hand side)
int g_currentX;
int g_currentY;

void createPlaceName(int x, int y);
void getName(int x, int y);

void init_token_buffers(void) {
  if (!g_tokenList) {
    g_tokenList = (u8*)malloc(256);
    g_tokenBuffer = (char*)malloc(32);
  }
  g_heap = g_tokenList;
}


// places ASCII values for token "tok" into g_bufferPos and updates g_tokenLength
// make sure g_bufferPos is valid before calling!!!!
void decode(int tok, const u8 * l_tokens) {
  int B = tok+1;
  int index = -1;
  do {
    index++;
    if ((l_tokens[index] & 0x20/*BIT05*/) == 0) {
      B--;
    }
  } while (B != 0);
  
  do {
    int data = l_tokens[index];
    char theChar = (char)((data|g_lowercase)&0xff);
    g_lowercase = 0x20;
    *g_bufferPos++ = theChar;
    g_tokenLength++;
    index++;
  } while ( (l_tokens[index] & 0x20) != 0);
  
}


// add the name of an area to heap (e.g. the Moon, Blood)
void addLiteralToken(char lit) {
  *g_heap++ = LITERAL;
  *g_heap++ = lit;
}

void addNewlineToken(void) {
  *g_heap++ = NEWLINE;
}


void addUppercaseToken(void) {
  *g_heap++ = UPPERCASE;
}

// pluralise if needed
void addPluralisedToken(u8 t) {
  *g_heap++ = t;
  if (g_isPlural)
    addLiteralToken('s');
}

void addSingleToken(u8 t) {
  *g_heap++ = t;
}


// copy a given token array to the heap
// copies until *from == ENDTOKEN
void addToTokenList(const u8 * from) 
{
  while (*from != ENDTOKEN)
    *g_heap++ = *from++;
}


void addPlaceToken(int x, int y) {
  *g_heap++ = PLACENAME;
  *g_heap++ = x&0xff;
  *g_heap++ = y&0xff;
}


// this might not be used...
int isTerrainPlural(int terrain) {
  return (terrain == MOUNTAINS ||
  terrain == WASTES ||
  terrain == PLAINS);
}


void addNameToken(int name) {
  *g_heap++ = UPPERCASE;
  *g_heap++ = name;
}

void addTokenNospace(int t) {
  *g_heap++ = NOSPACE;
  *g_heap++ = t;
}

//634d
void addDirection(int direction) {
  addNameToken(direction+0x18);
}



void makeUpperCase(char * theChar) {
  unsigned char c = (unsigned char)(*theChar);
  if (!g_capsLock) {
    if (!g_lowercase) {
      return;
    }
    g_lowercase = 0;
  }
  // got here? must make upper case.
  if (c >= (unsigned char)'a' && c <= (unsigned char)'z')
  {
    *theChar &= 0xdf;
  } 
}


// f6c0
// prints the chars in the 32 byte text buffer to the screen
void print_current_buffer(int palette) {
  int finalx = g_currentX + g_tokenLength;
  if (finalx >= g_maxX) {
    g_currentX = g_indent-1;
    g_currentY++;
  }
  g_currentX++;
  if (g_currentY == FINAL_ROW) { 
    // end of screen
    // wait for A press, clear text, reset position then carry on
    WaitKeyPress(KEY_A);
    clear_text_from(START_ROW);
    g_currentY = START_ROW;
    g_currentX = g_startX;
    // return;
  }
  // ok to draw text...
  // use void print_text(char text, int row, int col, int pal) in gfx.
  g_bufferPos = g_tokenBuffer;
  char theChar;
  do {
    theChar = *g_bufferPos++;
    // makeUpperCase(&theChar);
    print_text(theChar, g_currentX, g_currentY, palette);
    g_currentX++;
  }
  while (--g_tokenLength != 0);
  
}

// called during token expansion
void decodePlace(void) {
  
  g_heap++; //skip placename token
  
  // next 2 bytes are a place 
  int x = *g_heap++;
  int y = *g_heap++;
  createPlaceName(x,y);
  
}


// only call during token expansion
void addASCIIchar(void) {
  //f768  
  g_heap++; //skip literal token
  *g_bufferPos++ = *g_heap++;
  g_tokenLength++;
}

void tokenAndEnd(void) {
  decode(*g_heap, tokendict);
  g_heap++;
  while (*g_heap == LITERAL)  {
    addASCIIchar();
  }
  if (*g_heap == NOSPACE)  {
    g_heap++;
    tokenAndEnd();
  }
}

void print_tokens_newline_0indent(int palette) {
  print_tokens_pal(palette);
  g_indent = 0;
  addNewlineToken();
}


void set_print_vars(int x, int y, int indent, int maxx) {
  g_startX = x;
  g_startY = y;
  g_indent = indent;
  g_maxX = maxx;
  g_currentX = g_startX-1; 
  g_currentY = g_startY; 
}


// f7b5
void print_tokens(int x, int y, int indent, int maxx, int palette) {
  
  set_print_vars(x,y,indent,maxx);
  
  print_tokens_pal(palette);
}

void print_tokens_pal(int palette) {
  
  *g_heap++ = ENDTOKEN;
  
  // 6229 in ddr
  g_heap = g_tokenList;
  g_bufferPos = g_tokenBuffer;
  
  g_lowercase = 0x20;
  g_tokenLength = 0;
  
  while (*g_heap != ENDTOKEN) {
    if (*g_heap == LITERAL || *g_heap == PLACENAME)  {
      
      if (*g_heap == LITERAL)
        addASCIIchar();
      else 
        decodePlace();
      
      // jump to after decompress token
      while (*g_heap == LITERAL)  {
        addASCIIchar();
      }
      if (*g_heap == NOSPACE)  {
        g_heap++;
        tokenAndEnd();
      }
    } else if (*g_heap == UPPERCASE)  {
      g_heap++;
      g_lowercase = 0;
      // restart loop
      continue;
    } else if (*g_heap == NEWLINE)  {
      g_heap++;
      g_currentX = g_indent; 
      g_currentY++;
      // restart loop
      continue;
    } else {
      tokenAndEnd();
    }
    
    // f74a - finished decompressing, draw the text!
    print_current_buffer(palette);
    g_bufferPos = g_tokenBuffer;
    
  }  
  g_heap = g_tokenList;
}

/*
void setTens(int largeVal,
             int * units,
             int * tens,
             int * hundreds,
             int * thousands,
             int * tensthous
             ) {
  
  signed long result, modulus;
  FastIntDivide(largeVal, 10, &result, &modulus);
  *units = modulus;
  FastIntDivide(result, 10, &result, &modulus);

  *tens = modulus;
  FastIntDivide(result, 10, &result, &modulus);

  *hundreds = modulus;
  FastIntDivide(result, 10, &result, &modulus);

  *thousands = modulus;
  FastIntDivide(result, 10, &result, &modulus);

  *tensthous = modulus;
  
}



void addTeenEnding(int teen) {
  if (teen > 2) {
    if (teen == 8) {
      addToTokenList(een);
    } else {
      addTokenNospace(0xb8);
    }
  }
}


// code at e1aa

void  addNumber(int tens, int units) {
  if (tens == 0) {
    // one to ten
    *g_heap++ = units+0x62;
  } else if (tens == 1) {
    // e1eb
    // add the teens, 10-19
    *g_heap++ = theTeens[units];
    addTeenEnding(units);
  } else {
    *g_heap++ = theTees[tens-2];
    if (tens > 2) {
      if (tens != 8) {
        addLiteralToken('t');
      } 
      addLiteralToken('y');
    }
    
    if (units != 0) {
      addToTokenList(dashJoin);
      *g_heap++ = units+0x62;
    }
  }
}
*/

/*
  based on e21f
  Takes an int value and converts it to words 
  eg. 2001 -> "two thousand and one"
  This routine adds the tokens for the words to the token list
*/
/*
void numberToWords(int value) {
  
  
  //int index = 0;
  int units, tens, hundreds, thousands, tensthous;
  setTens(value, &units, &tens, &hundreds, &thousands, &tensthous);
  int total = units + tens + hundreds + thousands + tensthous;
  if (total == 0) {
    *g_heap++ = 0xbc;     // no
    return;
  }
  
  // check plural too...
  g_isPlural = 1;
  if (total == 1 && units == 1)
    g_isPlural = 0;
  int tensAndUnits = tens+units;
  int hundredsAndThousands = hundreds + thousands + tensthous;
  if (hundredsAndThousands != 0) {
    int justThousands = thousands + tensthous;
    if (justThousands != 0) {
      // e25b
      addNumber(tensthous, thousands);
      *g_heap++ = 0xbb; // thousand
    }
    if (hundreds != 0) {
      addNumber(0, hundreds);
      *g_heap++ = 0xba; // hundred
    }
    if (tensAndUnits  != 0) {
      *g_heap++ = 0x87; // and
    }  
  }
  
  if (tensAndUnits != 0) {
    addNumber(tens, units);
  }
  
}
*/

/*
  based on 7602, dr version
  Takes an int value and converts it to words 
  eg. 1951 -> "nineteen hundred, two score and eleven"
  This routine adds the tokens for the words to the token list
*/
void numberToWords(int n) {
  
  int needsComma = 1;
  int units = n%20;
  int hundreds = n/20; // not yet hundreds...
  
  // set is plural flag...
  g_isPlural = (units!=1);
  
  if (units == 0) {
    needsComma = 0;
  }
  
  int scores = hundreds%5;
  hundreds = hundreds/5; // is hundreds now
  
  if ( hundreds == 0) {
    needsComma = 0;
  }
  
  
  int needsAnd = 0;
  if ( hundreds!=0) { // has hundreds
    needsAnd = 1;
    *g_heap++ = 0x55+(hundreds);  // one, two, etc
    *g_heap++ = 0x53;  // hundred
  }
  if (scores != 0) {
    if ( needsComma != 0) {
      // add comma (had units or hundreds)
      addLiteralToken(',');
      *g_heap++ = 0x55+(scores);  // one, two, etc
        
    } else {
      if ( needsAnd != 0)  {
        *g_heap++ = 0x50;  // and
      }
      *g_heap++ = 0x55+(scores);  // one, two, etc           
      
    }
    *g_heap++ = 0x54;  // score
    needsAnd = 1;
  }
  
  if (units != 0) {
    if ( needsAnd != 0)  {
      *g_heap++ = 0x50;  // and
    }
    *g_heap++ = 0x55+(units);  // one, two, etc
    
    needsAnd = 1;
  }
  if ( needsAnd == 0) {
    *g_heap++ = 0x55;  // no
  }
  
}


// basic naming of an area. no tracing of "zones".
// name is in tokens, in the buffer
void getName(int x, int y) {
  int key = calculateNameSeed(x,y);
  int start = ((key&0x1f00)>>8);
  int middle = (key&0xf0)>>4;
  int end = key&0xf;
  g_lowercase = 0x00;
  decode(start,  &namedicts[NAME_DICT_1]);
  decode(middle, &namedicts[NAME_DICT_2]);
  decode(end,    &namedicts[NAME_DICT_3]);
  
}

// create a place name in the buffer.
// checks with zone tracing first
void createPlaceName(int x, int y) {

  if ((g_map[x+y*64]&0xf) < GATE) {
    // need to trace out path and stuff...
    // call 5f6c to get correct x, y
    location_t p = {x, y};
    traceArea(&p);
    x = p.x;
    y = p.y;
  }
  getName(x, y);
}

// "Luxor" , "Imgarorn", etc
void createCharacterName(int id) {
  addUppercaseToken();
  if (id <= SHARETH) {
    // hardcoded name
    addSingleToken(id+0x13);
  } else {
    // generated name
    addPlaceToken(g_fortresses[id].x, g_fortresses[id].y);
  }
}

// creates the character's title
// "Imgarorn the Giant"
// "Luxor the Moonprince", etc
void createCharacterTitle(int id) {
  createCharacterName(id);
  addToTokenList(the_);
  int race = g_characters[id].race_orders & 0xf;
  addSingleToken(raceNames[race]);
  
}

void createCharacterName_s(int id) {
  createCharacterName(id);
  addToTokenList(apostropheS);
}


// creates the tokens for:
// the Frozen Wastes OR
// the <terrain> of <place> (eg. the Plains of Farand)
void addTerrainToken(const location_t *pos) {
  addToTokenList(the_);
  
  locdat_t loc;
  loc.x = pos->x;
  loc.y = pos->y;
  
  fillLocationData(&loc);
  addSingleToken(loc.terrain);
  if (loc.terrain == WASTES) {
    // added "Frozen", now need "Wastes"
    addNameToken(loc.terrain+1);
  } else {
    // the Blah of Somewhere-or-other
    addToTokenList(of_);
    addPlaceToken(loc.x, loc.y);
  }
}


void addAtLocation(const location_t * pos) {
  g_locdat.x = pos->x;
  g_locdat.y = pos->y;
  
  fillLocationData(&g_locdat);
  *g_heap++ = phrasalEndings[g_locdat.terrain];
  
  addTerrainToken(pos);
  
}


void createHeStandsString(int id, int inTunnel) {
  createCharacterTitle(id);
  if (inTunnel)
  { // in tunnel
    addToTokenList(standsInTunnel);
    addDirection(LOOKING(g_characters[id].time_looking)); // north, etc
    addLiteralToken('.');
    addTimeTokens(TIME_LEFT(g_characters[id].time_looking));
  }
  else 
  {
    addSingleToken(0x2d); // stands
    
    // where? "on the plains of despair"
    addAtLocation(&g_characters[id].location);
    
    addSingleToken(0x2e); // looking
    
    addDirection(LOOKING(g_characters[id].time_looking)); // north, etc
    
    // where are we looking?
    locdat_t loc;
    loc.x = g_lookAtPos.x;
    loc.y = g_lookAtPos.y;
    
    fillLocationData(&loc);
    
    if (HAS_MIST(loc.info)) {
      addToTokenList(intoTheMist);
    } else {
      // are we looking at anything interesting?
      
      // if it is a gate
      // OR it is not the same thing that we are on
      // OR (deep breath) it is the same thing we are on, 
      // BUT neither are plains and the thing we are looking at 
      // is a way off, so is probably from a different realm anyway.
      if (loc.terrain >= GATE || loc.terrain != g_locdat.terrain ||
       (loc.terrain != PLAINS && g_plainsBetweenUs != 0)
      ) {
        addSingleToken(0x2c);
        addTerrainToken(&g_lookAtPos);
      } else {
        // meh. seen it, bought the t
        addLiteralToken('.');
      }
    }
  }
  
  
}


// time of day
// 767c
void addTimeTokens(int time) {
  addUppercaseToken();
  if (time == DAWN) {
    addToTokenList(itIsDawn);
  } else if (time == 0) {
    addToTokenList(itIsNight);
  } else {
    numberToWords( (time+3)/4 );
    addPluralisedToken(0x69); //hour(s)
    addToTokenList(ofTheDay);
    g_isPlural = !g_isPlural;
    addPluralisedToken(0x6b); //remain(s)
    addLiteralToken('.');
  }
  
}

//describe object ("the sword of Thigrim")
void addShortObjectDescription(int object) {
  
  addToTokenList(the_);
  // get object type, which is table 14. The same as character race!
  addSingleToken(objectTypes[g_characters[object].race_orders&0xf]);
  //objectUsees[g_characters[object].race_orders&0xf];
  addToTokenList(of_);
  //where is it of? the same place as the fortress!!
  addPlaceToken(g_fortresses[object].x,g_fortresses[object].y);
  
  
}
//; Full object description = ("the sword of Thigrim whose power is in battle")
void addFullObjectDescription(int object) {
  addShortObjectDescription(object);
  addToTokenList(whosePowerIsIn);
  addSingleToken(objectUses[g_characters[object].race_orders&0xf]);
  
}

void addUtterly(int status) {
  // 45 utterly
  // 46 extremely
  // 47 very
  // 48 quite
  // 49 somewhat
  // 4a slightly
  // 4b not
  // 4c not at all

  addSingleToken(0x45 + (status/32));
}

void addMoreAndPrint(void) {
  addToTokenList(_More);
  print_tokens_pal(FONT_RED);
}

const u8 loyaltyNames[] = {
  0x20,0x28,0x23,0x24,0x25,0x26
};


void addTitleAndLoyalty(int who) {
  createCharacterTitle(who);
  if (who != LUXOR && who != SHARETH) {
    addToTokenList(whoIsLoyalTo);
    addSingleToken(loyaltyNames[LOYALTY(g_characters[who].flags)]);
    if (LOYALTY(g_characters[who].flags)  >= 3) {
      addLiteralToken('s');
    }
  }
  addLiteralToken('.');
}
