/*
* $Log$
* Revision 2.2  2004/02/18 19:11:14  richq
* Very similar to the LoM token parsing, except the tokens aren't compressed into
* 5 bits and the general feel is "dumb and duller".
*  Number to words has changed, for the worse imo, as the "two score and three"
* thing is harder to follow.
*  The "utterly bold" type string is now not as complex (and always negative!)
*  Time now shows just full hours instead of the good old "less than..." thing
* Oh well...
*
* Revision 2.1.1.1  2004/02/04 21:49:20  Default
* Doomdark's Revenge GBA
*
*
*/

/*
* Everything to do with text tokens and token parsing
*/

#ifndef _TOKENS_H
#define _TOKENS_H

#include "gba.h"
#include "location.h"

extern int g_currentY;    // current row
extern int g_capsLock;
extern int g_isPlural;

extern int g_indent;

void init_token_buffers(void);


void set_print_vars(int x, int y, int indent, int maxx);
// print the tokens setting all the data for start position, indent, etc
void print_tokens(int x, int y, int indent, int maxx, int palette);
// print tokens using current settings, but change the palette only
void print_tokens_pal(int palette);
// 7aad ...
void print_tokens_newline_0indent(int palette);

/*Takes an int value and converts it to words 
  adds the tokens for the words to the token list */
void numberToWords(int value);
// adds an array of tokens to the list
void addToTokenList(const u8 * array);
// adds a single token:
void addSingleToken(u8 t);
// adds a pluralised token:
void addPluralisedToken(u8 t);
// capitalises a token, i.e. upper case then name:
void addNameToken(int name);
// adds just an uppercase token
void addUppercaseToken(void);
// adds a newline token
void addNewlineToken(void);
// adds a literal ASCII char
void addLiteralToken(char lit);
// adds the direction, north, etc
void addDirection(int direction);
// adds a placename token + the given position
void addPlaceToken(int x, int y);

// just the first name:
void createCharacterName(int id);
// the full title: Booboo the Wotsit, etc
void createCharacterTitle(int id);

// just the first name, with 's. e.g. Luxor's
void createCharacterName_s(int id);

void addTitleAndLoyalty(int who);

// creates the tokens for:
// the Frozen Wastes or
// the <terrain> of <place> (eg. the Plains of Farand)
void addTerrainToken(const location_t *pos);

//"on the <terrain> of <place>"
void addAtLocation(const location_t * pos);

void addDirection(int direction);

// "Luxor the Moonprince stands at the Gate of Viagra
// looking East to the Plains of Abracadabra"
void createHeStandsString(int id, int inTunnel);

// time of day, e.g. 4 hours of the day remain
void addTimeTokens(int time);

void addShortObjectDescription(int object);
void addFullObjectDescription(int object);

// not as good as in LoM, but adds the utterly->not at all status
void addUtterly(int status);
// adds "\nMore?" and prints in red
void addMoreAndPrint(void);


#endif

